local modName = "Encounters"
local parent = HudMap
local L = LibStub("AceLocale-3.0"):GetLocale("HudMap")
local mod = HudMap:NewModule(modName, "AceEvent-3.0")
local db

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")
local callbacks = {}

local options = {
	type = "group",
	name = L["Encounters"],
	args = {
		enable = {
			type = "toggle",
			name = L["Enable Encounters"],
			order = 100,
			get = function()
				return db.enable
			end,
			set = function(info, v)
				db.enable = v
				if v then
					mod:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")	
				else
					mod:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")	
				end
			end
		},
		zones = {
			type = "group",
			name = L["Zones"],
			args = {},
			get = function(info)
				local obj = db.zones
				for i = 3, #info do
					obj = obj[info[i]]
				end			
				if type(obj) == "table" then
					return obj.r, obj.g, obj.b, obj.a
				else
					return obj
				end
			end,
			
			set = function(info, ...)
				local obj = db.zones
				for i = 3, #info - 1 do
					obj = obj[info[i]]
				end
				if select("#", ...) == 1 then
					obj[info[#info]] = ...
				else
					local t = obj[info[#info]]
					t.r, t.g, t.b, t.a = ...
				end
			end			
		}
	}
}

local defaults = {
	profile = {
		enable = true,
		zones = {}		
	}
}

mod.group = HudMap.group

function mod:OnInitialize()
	self.db = parent.db:RegisterNamespace(modName, defaults)
	db = self.db.profile
	parent:RegisterModuleOptions(modName, options, modName)
end

function mod:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")	
	self:RegisterEvent("CHAT_MSG_RAID_BOSS_EMOTE", "Yell")
	self:RegisterEvent("CHAT_MSG_MONSTER_YELL", "Yell")
	self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE", "Yell")
	self:RegisterEvent("CHAT_MSG_MONSTER_SAY", "Yell")	
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(ev, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellID, ...)
	local ev = callbacks[event]
	if ev then
		for _, module in ipairs(ev) do
			module[event](module, spellID, sourceName, destName, sourceGUID, destGUID)
		end
	end
end

function mod:Yell(event, ...)
	local ev = callbacks[event]
	if ev then
		for _, module in ipairs(ev) do
			module[event](module, ...)
		end
	end
end

do
	local function option(self, setting)
		local t = db.zones[self.zone:gsub(" ", "")][self.name:gsub(" ", "")][setting]	
		if type(t) == "table" then
			return t.r, t.g, t.b, t.a
		else
			return t
		end
	end

	function mod:RegisterModule(zone, ...)	
		for i = 1, select("#", ...) do
			local module = select(i, ...)
			if module.name and module.options then
				options.args.zones.args[zone:gsub(" ", "")] = options.args.zones.args[zone:gsub(" ", "")] or {
					type = "group",
					name = zone,
					args = {}
				}
				module.Option = option
				module.zone = zone:gsub(" ", "")
				local t = options.args.zones.args[zone:gsub(" ", "")]
				if module.defaults then
					local gzone = zone:gsub(" ", "")
					defaults.profile.zones[gzone] = defaults.profile.zones[gzone] or {}
					defaults.profile.zones[gzone][module.name:gsub(" ", "")] = module.defaults
				end
				t.args[module.name:gsub(" ", "")] = {
					type = "group",
					name = module.name,
					args = module.options
				}
			end
			if module.registerOn then
				for _, e in ipairs(module.registerOn) do
					callbacks[e] = callbacks[e] or {}
					tinsert(callbacks[e], module)
				end
			end
		end
	end
end

local approximateUnits = {}
function mod:ApproximateBossPosition()
	wipe(approximateUnits)
	local prefix, max
	if GetNumRaidMembers() > 0 then
		prefix = "party"
		max = "4"
	elseif GetNumPartyMembers() > 0 then		
		prefix = "raid"
		max = 25
	end
	for i = 1, max do
		local _, cls = UnitClass(prefix..i)
		if not cls then break end
		
		if cls == "HUNTER" and UnitExists(prefix..i.."pet") then
			tinsert(approximateUnits, prefix..i)
		elseif cls == "ROGUE" or cls == "WARRIOR" or cls == "DEATHKNIGHT" then
			tinsert(approximateUnits, prefix..i)
		end
	end
	local sumX, sumY, sumN = 0, 0, 0
	for _, v in ipairs(approximateUnits) do
		local x, y = HudMap:GetUnitPosition(v)
		if x and y and x ~= 0 and y ~= 0 then
			sumX = sumX + x
			sumY = sumY + y
			sumN = sumN + 1
		end
	end
	return sumX / sumN, sumY / sumN
end

mod.GuidToMobID = setmetatable({}, {__index = function(t, k)
	local id = (k and tonumber(k:sub(9, 12), 16)) or 0
	rawset(t, k, id)
	return id	
end})

function mod:IsHeroic()
	local diff, switchable = select(6, GetInstanceInfo())
	if switchable then
		return diff == 1
	else
		return GetInstanceDifficulty() >= 3
	end
end

do
	function mod:RangeCheckAll(tex, radius, duration, nr, ng, nb, na, excludePlayer)
		for index, unit in mod.group() do
			local n = UnitName(unit)
			if n and not UnitIsDead(unit) and not UnitIsUnit("player", unit) then
				local c = HudMap:PlaceRangeMarkerOnPartyMember(tex, n, radius, duration, nr, ng, nb, na):Appear():Rotate(360, duration):RegisterForAlerts(true)
			end
		end	
	end
end

function mod:HomeToTarget(unit, radius, r, g, b, a)
	radius = radius or 2
	r = r or 1
	g = g or 1
	b = b or 1
	a = a or 1	
end

function mod:GetMobTarget(mobID)
	for index, unit in mod.group() do
		local g = UnitGUID(unit.."target")
		if g then
			if mod.GuidToMobID[g] == mobID then
				return UnitName(unit.."targettarget")
			end
		end
	end	
end

do
	local ABS, POW = math.abs, math.pow
	local gpu = HudMap.GetUnitPosition
	function mod:GetCenterOfRaid(tolerance)
		local sx, sy, sc = 0, 0, 0
		for index, unit in mod.group() do
			local x, y = gpu(HudMap, unit)
			sx = sx + x
			sy = sy + y
			sc = sc + 1
		end
		local ax, ay = sx / sc, sy / sc
		sx, sy, sc = 0, 0, 0
		for index, unit in mod.group() do
			local x, y = gpu(HudMap, unit)
			local dx = ax - x
			local dy = ay - y
			if ABS(POW((dx*dx)+(dy*dy), 0.5)) < tolerance then
				sx = sx + x
				sy = sy + y
				sc = sc + 1
			end
		end
		return sx / sc, sy / sc
	end
end

do
	local delays = {}
	function mod:Delay(func, delay)
		delays[func] = GetTime() + delay
	end

	local timerFrame = CreateFrame("Frame", nil, UIParent)
	timerFrame:SetWidth(1)
	timerFrame:SetHeight(1)
	timerFrame:SetScript("OnUpdate", function(self, t)
		local ct = GetTime()
		for k, v in pairs(delays) do
			if ct > v then
				k()
				delays[k] = nil
			end
		end
	end)
end